# Web-based visualization of SPIR-V: getting started

This project is about creating a visualization tool for
SPIR-V programs, inspired by
[what the Go compiler can do](https://pauladamsmith.com/blog/2016/08/go-1.7-ssa.html):

![Go visualization example](https://pauladamsmith.com/images/gossa/thumb.png)

Unlike the example above,
the control-flow graph should probably be represented,
mostly likely as an actual graph with nodes and edges.

[SPIR-V](https://www.khronos.org/registry/spir-v/specs/unified1/SPIRV.pdf) is
an intermediate representation for programming GPUs,
along the lines of LLVM IR.

## How this project could be implemented

There is a lot of flexibility in
how to implement this project,
as it is only loosely linked to
the
[main GraphicsFuzz project](https://github.com/google/graphicsfuzz).
You will most likely be developing a standalone tool/viewer.
The main set of tools for working with SPIR-V
are developed under the
[SPIRV-Tools repository](https://github.com/KhronosGroup/SPIRV-Tools).
You may want to
make use of this code in some way via C++,
but the amount of C++ code you write could be
small (including none at all)
or large depending on your approach.

One approach is to create a new tool inside SPIRV-Tools
that outputs a self-contained HTML file.
Another approach would be to output just some
data (e.g. in JSON or protobuf format)
that is then processed by some other command line tool
or web app/page.
Or, you could create a web app/page
that does some clever stuff (e.g. using emscripten)
to directly use the SPIRV-Tools code
via JavaScript/TypeScript so that the user can
just drag-and-drop a SPIR-V file into your web app/page.
Alternatively,
your tool could just process the disassembled SPIR-V files
that look similar to this:

```
       OpCapability Shader
  %1 = OpExtInstImport "GLSL.std.450"
       OpMemoryModel Logical GLSL450
       OpEntryPoint Fragment %4 "main"
       OpExecutionMode %4 OriginUpperLeft
       OpSource ESSL 310
       OpName %4 "main"
       OpName %8 "x"
  %2 = OpTypeVoid
  %3 = OpTypeFunction %2
  %6 = OpTypeInt 32 1
  %7 = OpTypePointer Function %6
  %9 = OpConstant %6 1
 %10 = OpConstant %6 2
 %11 = OpConstant %6 3
 %12 = OpConstant %6 4
  %4 = OpFunction %2 None %3   ; the start of a function
  %5 = OpLabel
  %8 = OpVariable %7 Function
       OpBranch %13
 %13 = OpLabel           ; OpLabel denotes the start of a block
       OpStore %8 %9
       OpBranch %14      ; branches are also easy to identify
 %14 = OpLabel
       OpStore %8 %10
       OpBranch %15
 %15 = OpLabel
       OpStore %8 %11
       OpBranch %16
 %16 = OpLabel
       OpStore %8 %12
       OpBranch %17
 %17 = OpLabel
       OpReturn       ; this block terminates with a return
       OpFunctionEnd  ; end of function
```


The code for this project will likely reside
in the
[main GraphicsFuzz repository](https://github.com/google/graphicsfuzz)
or a new repository,
possibly with some additional code
in the
[SPIRV-Tools repository](https://github.com/KhronosGroup/SPIRV-Tools).

## Getting started

To get a taste of what the project will entail,
we recommend keen applicants to at least try
the following.
Make sure you read the last paragraph
of this section where we note
an alternative approach that does not require using C++.
Also see the [Viewer](#viewer) section
for ideas for prototyping the viewer application itself.

* "Git clone" and build (using CMake) the following projects:
https://github.com/KhronosGroup/glslang
https://github.com/KhronosGroup/SPIRV-Tools

You should then have the following command line tools available: `glslangValidator`, `spirv-dis`, `spirv-as`.

* Download the following sample
GLSL fragment shader:
https://github.com/SaschaWillems/Vulkan/blob/master/data/shaders/inlineuniformblocks/pbr.frag

It is not that interesting in terms of control flow, but you can always edit it
to add some if statements and loops.

* Then do:

```sh
# Translate the GLSL fragment shader to a SPIR-V binary
glslangValidator -V pbr.frag -o pbr.frag.spv
# Disassemble it
spirv-dis pbr.frag.spv -o pbr.frag.asm --raw-id
```

You now have an example of a SPIR-V binary `pbr.frag.spv`, as well a disassembled version `pbr.frag.asm` (that you can open in a text editor).

* "Git clone" this fork of SPIRV-Tools
(or add it as a remote, and fetch): 
https://github.com/paulthomson/SPIRV-Tools

* Git checkout the `vis` branch.
This branch contains the skeleton of a new tool
`spirv-vis`. The code is here:
https://github.com/paulthomson/SPIRV-Tools/blob/vis/tools/vis/vis.cpp

* Build the project (again, using CMake)
and then do:

`spirv-vis pbr.frag.spv`

The file `output.txt` will be created and should contain
the ID of every function. E.g.

```
Function 4
Function 12
Function 18
...
```

Confirm this by opening `pbr.frag.asm`
and searching for e.g. `%4 = `;
you should find that ID 4 is indeed a function:

```
%4 = OpFunction %2 None %3
```

* You should try modifying `vis.cpp` to output more details.
For example:
  * output every block of a function (you can iterate over these in a similar way to iterating over the functions in a module)
  * output every instruction of every block (again, iteration is similar here)
  * output the capabilities,
extensions,
memory model,
types,
constants,
and global variables (see `source/opt/module.h`)
  * output every usage of every instruction using the def-use manager (e.g. `Instruction defined ID %72, which is used in instructions: %74`)
  * output the successors of every block
  * output the merge block of every block, if any

* If you want to show us your prototype,
feel free to share your repository with us.

Note that in the `vis` skeleton tool,
we are using the `spirv-opt` classes
to enumerate features of the SPIR-V module.
However,
this is just one possible approach.
Notice that the disassembled file `pbr.frag.asm`
is already fairly informative:
IDs are defined via `%X = `
and used elsewhere via `%X`;
blocks start with a label `%X = OpLabel`;
branches can be identified via the instruction name
(e.g. `OpBranchConditional %165 %166 %167`)
and the targets are given by certain parameters
(e.g. `%166 %167`).
See the [SPIR-V spec](https://www.khronos.org/registry/spir-v/specs/unified1/SPIRV.pdf).
Thus, you may be able to create a visualization
straight from the `.asm` file.
See the [Viewer](#viewer) section below for ideas for prototyping
the actual viewer, which could include the code for processing
`.asm` files.


## spirv-cfg

The SPIRV-Tools project already has a `spirv-cfg` tool.
Try it by doing the following:

```sh
spirv-cfg pbr.frag.spv -o cfg.dot

# Use any dot file viewer.
# E.g., on most Linux distributions:
dot -Tx11 cfg.dot
```

You will see the control flow graph for each function,
plus the
merge blocks
(to see these,
try creating if statements, switch statements, and loops
in the original GLSL shader)
and continue targets
(to see these, create loops in the original GLSL shader).

Your tool should improve upon this basic tool by supporting some or all
of the following features:

* Show all instructions, not just blocks.
* Interactive graph;
drag and drop nodes, click on IDs to highlight their uses, etc.
* Loop constructs (e.g. for loops)
and selection constructs (i.e. if and switch statements)
have a **header block** and a **merge block**,
where the only entry into the construct is via the header block
and most exits from the construct are via the merge block.
The nodes should be organized accordingly;
e.g. the "loop body" should appear between the header
and merge blocks,
while the instructions that come after the loop
should appear below the merge block;
this is not always the case with most graph layout algorithms,
especially if the loop body has more statements than the code that follows.
* A diff view to compare two related SPIR-V files.

## Viewer

For the actual viewer,
we anticipate that you will want to build a web app or generate a web page,
and use some
existing TypeScript/JavaScript libraries,
such as graph and/or graphics libraries.
This ensures the viewer will run on various platforms
with minimal effort,
and will probably be fairly consistent.
There are many different ways to create web apps/pages,
so there is flexibility in how you approach this. 

Please discuss this with us if you have ideas or suggestions.

We encourage keen applicants to create a small prototype
web app, web page, or other type of app,
demonstrating how they will visualize
the control-flow graphs of a SPIR-V file.
If you want to show us your prototype,
feel free to send screenshots with descriptions or
possibly a
link to your repository.
We are not experts in this domain (GUIs, web apps, etc.),
so any evidence that you have a plan for this
is very useful.


